package tools.descartes.prisma.core.agent;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

import org.eclipse.emf.cdo.CDOAdapter;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.notify.Notifier;
import org.eclipse.emf.ecore.EObject;

import tools.descartes.prisma.core.ModelRepository;
import tools.descartes.prisma.core.Transaction;
import tools.descartes.prisma.core.adapter.ModelSkeletonAdapter;
import tools.descartes.prisma.core.adapter.ModelSkeletonDelta;
import tools.descartes.prisma.core.exceptions.CommitException;
import tools.descartes.prisma.core.scopes.AgentScope;
import tools.descartes.prisma.core.scopes.SystemGlobalScope;
import tools.descartes.prisma.model.scope.Agent;
import tools.descartes.prisma.model.scope.DelegatedObject;
import tools.descartes.prisma.model.scope.ScopeConfiguration;
import tools.descartes.prisma.model.scope.ScopePackage;
import tools.descartes.prisma.model.skeleton.ModelSkeleton;

public class AgentController {

	private class DelegationRepositoryAdapter implements CDOAdapter {

		private Agent target;

		@Override
		public void notifyChanged(Notification msg) {
			if (msg.getFeatureID(
					ScopeConfiguration.class) == ScopePackage.SCOPE_CONFIGURATION__DELEGATED_OBJECTS) {
				if (msg.getEventType() == Notification.ADD) {
					DelegatedObject delegatedElement = (DelegatedObject) msg.getNewValue();
					notifyAddDelegatedElement(delegatedElement.getObject(),
							delegatedElement.getAdditionalProperties().map());
				} else if (msg.getEventType() == Notification.ADD_MANY) {
					List<?> newValues = (List<?>) msg.getNewValue();
					for (Object curValue : newValues) {
						DelegatedObject delegatedElement = (DelegatedObject) curValue;
						notifyAddDelegatedElement(delegatedElement.getObject(),
								delegatedElement.getAdditionalProperties().map());
					}
				} else if (msg.getEventType() == Notification.REMOVE) {
					DelegatedObject delegatedElement = (DelegatedObject) msg.getOldValue();
					notifyRemoveDelegatedElement(delegatedElement.getObject(),
							delegatedElement.getAdditionalProperties().map());
				} else if (msg.getEventType() == Notification.REMOVE_MANY) {
					List<?> oldValues = (List<?>) msg.getOldValue();
					for (Object curValue : oldValues) {
						DelegatedObject delegatedElement = (DelegatedObject) curValue;
						notifyRemoveDelegatedElement(delegatedElement.getObject(),
								delegatedElement.getAdditionalProperties().map());
					}
				}
			}
		}

		@Override
		public Agent getTarget() {
			return target;
		}

		@Override
		public void setTarget(Notifier newTarget) {
			target = (Agent) target;
		}

		@Override
		public boolean isAdapterForType(Object type) {
			return (type instanceof Agent);
		}
	}

	private Agent agent;
	private ModelRepository repository;
	private final ScheduledExecutorService workerThread = Executors.newScheduledThreadPool(1);
	private final String scopePath;
	private final String id;
	private AgentScope scope;

	private DelegationProvider delegationProvider = new DelegationProvider() {
		@Override
		public boolean needsDelegation(EObject object) {
			return false;
		}

		@Override
		public Map<String, String> getAdditionalProperties(EObject object) {
			return Collections.emptyMap();
		}
	};

	private volatile boolean running = false;

	public AgentController(ModelRepository repository, String scopePath, String id) {
		this.repository = repository;
		this.scopePath = scopePath;
		this.id = id;
	}

	public Agent getAgent() {
		return agent;
	}

	public AgentScope getContainerScope() {
		return scope;
	}

	public void setDelegationProvider(DelegationProvider provider) {
		this.delegationProvider = provider;
	}

	public Future<?> start() {
		return workerThread.submit(new Runnable() {
			public void run() {
				if (!running) {
					startAsync();
				}
				running = true;
			}
		});
	}

	public Future<?> stop() {
		return workerThread.submit(new Runnable() {
			public void run() {
				if (running) {
					stopAsync();
				}
				running = false;
			}
		});
	}

	public void apply(ModelSkeleton skeleton) {
		try (Transaction transaction = getModelRepository().createTransaction()) {
			scope.synchronizeModelSkeleton(transaction, skeleton);

			boolean success = true;

			try {
				transaction.commit();
			} catch (CommitException ex) {
				ex.printStackTrace();
				success = false;
			}

			List<ModelSkeletonAdapter> adapters = new LinkedList<>();
			for (EObject curObj : skeleton.eContents()) {
				ModelSkeletonAdapter curAdapter = getModelRepository().adapt(curObj, ModelSkeletonAdapter.class);
				adapters.add(curAdapter);
			}
			// IMPORTANT: commit after all model skeleton adapters have been
			// synchronized and merged.
			for (ModelSkeletonAdapter curAdapter : adapters) {
				ModelSkeletonDelta delta = curAdapter.commitFinished(success);

				// Delegate the new objects.
				for (EObject curObject : delta.getNewObjects()) {
					if (delegationProvider.needsDelegation(curObject)) {
						scope.delegate(curObject, delegationProvider.getAdditionalProperties(curObject));
					}
				}
				// Undelegate the removed objects.
				for (EObject curObject : delta.getRemovedObjects()) {
					if (delegationProvider.needsDelegation(curObject)) {
						// TODO:
					}
				}
			}
		}
	}

	protected ScheduledFuture<?> schedule(Runnable command, long period, TimeUnit unit) {
		return workerThread.scheduleAtFixedRate(command, 0, period, unit);
	}

	protected void startAsync() {
		repository.connect();
		SystemGlobalScope systemScope = getModelRepository().getSystemScope();
		systemScope.create();
		this.scope = (AgentScope) getModelRepository().getScope(scopePath);
		this.scope.create();
		this.agent = scope.addAgent(id);

		// Listen for changes in the delegated elements list in our scope so
		// that we can react to delegation requests we are interested in.
		this.scope.getScopeConfiguration().eAdapters().add(new DelegationRepositoryAdapter());
	}

	protected void stopAsync() {
		try {
			scope.removeAgent(agent);
		} finally {
			repository.disconnect();
		}
	}

	public ModelRepository getModelRepository() {
		return repository;
	}

	protected void notifyAddDelegatedElement(EObject newElement, Map<String, String> additionalProperties) {
	}

	protected void notifyRemoveDelegatedElement(EObject oldElement, Map<String, String> additionalProperties) {
	}

	protected void checkRunning() {
		if (!running) {
			throw new IllegalStateException("The agent is not running.");
		}
	}
}
