package tools.descartes.prisma.core.merge;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.Set;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;

import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.AbstractAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.BranchAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.BranchProbabilities;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ComponentInternalBehavior;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ServicebehaviorFactory;
import tools.descartes.prisma.core.ModelRepository;
import tools.descartes.prisma.core.Transaction;
import tools.descartes.prisma.core.adapter.ObjectMatcher;

public class ComponentInternalBehaviorMerger extends ObjectMerger {

	private static final String MERGE_IDENTIFIER = "[[PRISMA_MERGE]]";

	public ComponentInternalBehaviorMerger(ModelRepository repository) {
		super(repository);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * tools.descartes.prisma.core.merge.ObjectMerger#mergeTree(tools.descartes.
	 * prisma.core.Transaction, org.eclipse.emf.ecore.EObject,
	 * org.eclipse.emf.ecore.EObject, java.util.Set)
	 */
	@Override
	public void mergeTree(Transaction transaction, EObject base, EObject changed,
			Set<EStructuralFeature> changedFeatures) {
		// A special merging strategy is required for component internal
		// behaviors, since
		// they can contain an ordered list of actions. The ordering of these
		// actions need to
		// be preserved when merging two independently create component internal
		// behaviors
		ComponentInternalBehavior baseBehavior = (ComponentInternalBehavior) base;
		ComponentInternalBehavior changedBehavior = (ComponentInternalBehavior) changed;
		if (changedBehavior.getActions().isEmpty()) {
			if (baseBehavior.getActions().isEmpty()) {
				return;
			} else {
				if (!isMergeBranch(baseBehavior.getActions().get(0))) {
					insertAlternativeBranch(baseBehavior, 0, baseBehavior.getActions().size(),
							Collections.<AbstractAction> emptyList());
				} else {
					// TODO
				}
				return;
			}
		} else {
			int lastMatch = -1;
			Queue<AbstractAction> actionsToMerge = new LinkedList<>(changedBehavior.getActions());
			while (!actionsToMerge.isEmpty()) {
				AbstractAction curAction = actionsToMerge.peek();
				int match = findMatch(baseBehavior, lastMatch, curAction);
				if (match == (lastMatch + 1)) {
					if (match > (lastMatch + 1)) {
						if (baseBehavior.getActions().isEmpty() || !isMergeBranch(baseBehavior.getActions().get(lastMatch + 1))) {
							insertAlternativeBranch(baseBehavior, (lastMatch + 1), match,
									Collections.<AbstractAction> emptyList());
						} else {
							// TODO
						}
					}
					actionsToMerge.poll();
					lastMatch = match;
				} else {
					List<AbstractAction> conflictingActions = new LinkedList<>();
					int returnIdx = determineConflictingActions(baseBehavior, lastMatch, conflictingActions, actionsToMerge);
					boolean newMergeBranch = true;
					if (baseBehavior.getActions().isEmpty() || !isMergeBranch(baseBehavior.getActions().get(lastMatch + 1))) {
						newMergeBranch = true;
					} else {
						if (returnIdx == (lastMatch + 2)
								|| (returnIdx < -1 && baseBehavior.getActions().size() == (lastMatch + 2))) {
							newMergeBranch = false;
						}
					}

					if (newMergeBranch) {
						int end = (returnIdx < 0) ? baseBehavior.getActions().size() : returnIdx;
						insertAlternativeBranch(baseBehavior, lastMatch + 1, end, conflictingActions);
						lastMatch = returnIdx;
					} else {
						BranchAction mergeBranch = (BranchAction) baseBehavior.getActions().get(lastMatch + 1);
						ComponentInternalBehavior existingAlternative = null;
						for (ComponentInternalBehavior curBranch : mergeBranch.getBranches()) {
							if (!curBranch.getActions().isEmpty()) {
								if (ObjectMatcher.matches(curBranch.getActions().get(0), curAction)) {
									existingAlternative = curBranch;
									break;
								}
							}
						}
						if (existingAlternative == null) {
							existingAlternative = ServicebehaviorFactory.eINSTANCE.createComponentInternalBehavior();
							existingAlternative.getActions().addAll(conflictingActions);
							mergeBranch.getBranches().add(existingAlternative);
						} else {
							// TODO
						}
					}
				}
			}
		}
	}

	private void insertAlternativeBranch(ComponentInternalBehavior base, int conflictStart, int conflictEnd,
			List<AbstractAction> conflictingActions) {
		ComponentInternalBehavior newAlternative = ServicebehaviorFactory.eINSTANCE.createComponentInternalBehavior();
		newAlternative.getActions().addAll(conflictingActions);
		ComponentInternalBehavior existingAlternative = ServicebehaviorFactory.eINSTANCE
				.createComponentInternalBehavior();
		for (int i = conflictStart; i < conflictEnd; i++) {
			existingAlternative.getActions().add(base.getActions().get(conflictStart));
		}
		BranchAction mergeAction = insertMergeBranch(base, conflictStart);
		mergeAction.getBranches().add(existingAlternative);
		mergeAction.getBranches().add(newAlternative);
	}

	private int determineConflictingActions(ComponentInternalBehavior base, int basePosition,
			List<AbstractAction> conflictingActions, Queue<AbstractAction> actionsToMerge) {
		while (!actionsToMerge.isEmpty()) {
			AbstractAction curAction = actionsToMerge.poll();
			int idx = findMatch(base, basePosition, curAction);
			if (idx >= 0) {
				return idx;
			} else {
				conflictingActions.add(curAction);
			}
		}
		return -1;
	}

	private int findMatch(ComponentInternalBehavior behavior, int startIdx, AbstractAction searchedAction) {
		int numActions = behavior.getActions().size();
		for (int i = (startIdx + 1); i < numActions; i++) {
			if (ObjectMatcher.matches(behavior.getActions().get(i), searchedAction)) {
				return i;
			}
		}
		return -1;
	}

	private boolean isMergeBranch(AbstractAction action) {
		if (action != null) {
			if (action instanceof BranchAction) {
				return ((BranchAction) action).getBranchProbabilities().getName().startsWith(MERGE_IDENTIFIER);
			}
		}
		return false;
	}

	private BranchAction insertMergeBranch(ComponentInternalBehavior behavior, int position) {
		AbstractAction action = null;
		if (behavior.getActions().size() > position) {
			action = behavior.getActions().get(position);
		}
		BranchAction mergeBranch;
		if (isMergeBranch(action)) {
			mergeBranch = (BranchAction) action;
		} else {
			mergeBranch = ServicebehaviorFactory.eINSTANCE.createBranchAction();
			BranchProbabilities probabilities = ServicebehaviorFactory.eINSTANCE.createBranchProbabilities();
			probabilities.setName(MERGE_IDENTIFIER);
			mergeBranch.setBranchProbabilities(probabilities);
			behavior.getActions().add(position, mergeBranch);
		}
		return mergeBranch;
	}

}
